-------------------------------------------------------
-- Worker of Biped v0.74
--
--  Created by Dennis Roche
--   godennis@gmail.com
--
--  Inspired by Bruce Creevey
--
--  Based on Biped Worker by
--   David Humpherys
--
--  A set of tools to make selecting bipeds easier.
--  It supports several Bipeds with different bone
--  setups. Multi-bone (Control-Click), hierarchical
--  (Shift-Click), symmetry and opposite selections
--  are also supported, as are the setting of commonly
--  use pivot points all within the GUI.
--
--  Requires Adobe Flash Player 8
--  http://www.adobe.com/go/getflashplayer
--
--  Any feature requests please e-mail me.
--
-------------------------------------------------------

-------------------------------------------------------
-- START
-------------------------------------------------------

global gScriptRootDir = getDir #scripts + "\\WorkerOfBiped\\"
global gBipedNode = undefined

global ui_bipedList
global ui_selectionTools
global ui_bipedControl
global ui_propertyLibrary

-------------------------------------------------------
-- STRUCTS
-------------------------------------------------------

struct libraryData(name="", tension=undefined, continuity=undefined, bias=undefined, easeTo=undefined, easeFrom=undefined, ikBlend=undefined, ikSpace=undefined, ikJoinedPivot=undefined)
struct bipSelection(index=undefined, link=undefined)
struct pivotSelection(index=undefined, link=undefined, pivot=undefined)

-------------------------------------------------------
-- LOGIC
-------------------------------------------------------


/*
 * function: getBipedsInScene
 *  finds all the bipeds in the scene and
 *   return an array
 *
 */
fn getBipedsInScene =
(
	-- gather bipeds in scene
	local bipedArray = #()
	for g in geometry where classof g.baseobject == Biped_Object do
	(
		if g.parent != undefined then (
			if classof g.baseobject != Biped_Object then (
				append bipedArray g.name
			) else if classof g.parent == Dummy then (
				append bipedArray g.name
			)
		) else (
			append bipedArray g.name
		)
	)


	return bipedArray
)


/*
 * function: getBipParent
 *  recursive function
 *  find the bip part parent
 *
 */
fn getBipParent bipPart =
(
	if bipPart.transform.isanimated == false then (
		if bipPart.parent != undefined then (
			getBipParent bipPart.parent
		)
	) else (
		return bipPart
	)
)


/*
 * function: setKey
 *  change key
 *
 */
fn setBipKey key param val =
(
	-- set key
	case param of
	(
		#tension:( key.tension = val )
		#continuity: ( key.continuity = val )
		#bias:( key.bias = val )
		#easeTo:( key.easeTo = val )
		#easeFrom:( key.easeFrom = val )
		#ikBlend:( key.ikblend = val)
		#ikSpace: (key.ikspace = val)
		#ikJoinedPivot:(
			if val > 0 then key.ikJoinedPivot = true
			else key.ikJoinedPivot = false
		)
	)
)


/*
 * function: changeKeyParam
 *  changes the param to val at keytime
 *
 */
fn changeKeyParam param val keytime =
(
	-- with anything selected
	for obj in selection do (
		local bipPart = getBipParent(obj)
		case classof bipPart.transform.controller of
		(
			-- BipSlave_Control
			-- Matrix3Controller
			-- the transform controller for each object
			BipSlave_Control:
			(
				-- for all keys in this controller
				for i = 1 to bipPart.transform.controller.keys.count do (
					-- get a key
					k = biped.getkey bipPart.transform.controller i
					-- set key
					if keytime == undefined then (
						setBipKey k param val
					) else if keytime == k.time then(
						setBipKey k param val
					)
				)
			)

			-- Vertical_Horizontal_Turn
			-- Matrix3Controller
			-- transform controller for the root object
			Vertical_Horizontal_Turn:
			(
				local track = bipPart.transform.controller.trackselection

				-- why swap the tracks around??
				if track == 1 then track = 2
				if track == 2 then track = 1

				local debug = newScript()
				
				format "track -> %\n" track to:debug
				format "bipPart -> %\n" bipPart to:debug
				format "bipPart.transform -> %\n" bipPart.transform to:debug
				format "bipPart.transform.controller -> %\n" bipPart.transform.controller to:debug
				format "bipPart.transform.controller[track] -> %\n" bipPart.transform.controller[track] to:debug
				
				-- for all keys in this controller
				for i = 1 to bipPart.transform.controller.keys.count do (
					-- get a key
					k = biped.getkey bipPart.transform.controller i
					-- set key
					if keytime == undefined then (
						setBipKey k param val
					) else if keytime == k.time then(
						setBipKey k param val
					)
				)
			)

		)
	)
)

/*
 * function: deleteBipKeyAtCurrentTime
 *  deletes the keys at current time
 *
 */
function deleteBipKey bipPart =
(
	-- if NOT animated
	if bipPart.transform.isanimated == false then
	(
		if bipPart.parent != undefined then (
			deleteBipKey bipPart.parent
		)
	) else (
		-- for all keys in this controller
		local c = 1
		for k in bipPart.transform.controller.keys do
		(
			if k.time == slidertime then
			(
				print "deleting key"
				deletekey BipPart.transform.controller c
			)
			c += 1
		)
	)
)


/*
 * function: getAllBipedNodes
 *  returns an array of all biped nodes
 *
 */
fn getAllBipedNodes =
(
	-- we want get all the biped nodes
	-- ignoring footfalls and props
	-- node indexes: 1-15, 17-19
	local bipedNodes = #()

	for i = 1 to 19 where i != 16 do (
		for j = 1 to 15 do (
			try (
				local node = (biped.getNode gBipedNode i link:j)
				if node != undefined then (
					append bipedNodes node
				)
			) catch ()
		)
	)

	return bipedNodes
)


/*
 * function: getBipedLinks
 *  returns an array of all biped nodes
 *
 */
fn getBipedLinks index =
(
	-- we want to get the links from
	-- the supplied biped index
	local bipedNodes = #()

	if (biped.getNode gBipedNode index) != undefined then (
		for j = 1 to 15 do (
			try (
				local node = (biped.getNode gBipedNode index link:j)
				if node != undefined then (
					append bipedNodes node
				)
			) catch ()
		)
	)

	return bipedNodes
)


/*
 * function: selectRoot
 *  selects the root node of the biped
 *
 */
fn selectRoot =
(
	if IsValidNode(gBipedNode) then (
		-- select biped root node
		select(gBipedNode)
	)
)


/*
 * function: selectBody
 *  selects the entire biped body
 *   ignores footfalls, props, and arm twist nodes
 */
fn selectBody =
(
	if IsValidNode(gBipedNode) then (
		clearNodeSelection()
		-- select all nodes
		local nodes = getAllBipedNodes()
		selectMore nodes
	)
)


-------------------------------------------------------
-- FLASH ACTIVE X
-------------------------------------------------------


/*
 * function: flash_FSCommand
 *  process the FSCommand sent from the
 *  flash activex object
 *
 */
fn flash_FSCommand cmd args =
(
	--format "flash_FSCommand -> % %\n" cmd args
	case cmd of (

		"clearSelection":
			-- clear selection
			clearNodeSelection()

		"selectMore":
			-- select more
			try (
				-- read and evaluate as an expression
				-- bipSelection index link
				local bs = readExpr (args as stringstream)
				if bs.link != 0 then (
					selectMore(biped.getNode gBipedNode bs.index link:bs.link)
				) else (
					selectMore(biped.getNode gBipedNode bs.index)
				)
			) catch ()

		"selectPivot":
			-- select pivot
			try (
				-- read and evaluate as an expression
				-- pivotSelection index link
				local ps = readExpr (args as stringstream)
				select (biped.getNode gBipedNode ps.index link:ps.link)
				local keyIndex = getKeyIndex ($.controller) currentTime
		        if keyIndex != 0 then (
		            key = biped.getKey $.controller keyIndex
		            key.ikPivotIndex = ps.pivot
		        )

			) catch ()


		"deselect":
			-- select more
			try (
				-- read and evaluate as an expression
				-- bipSelection index link
				local bs = readExpr (args as stringstream)
				if bs.link != 0 then (
					deselect(biped.getNode gBipedNode bs.index link:bs.link)
				) else (
					deselect(biped.getNode gBipedNode bs.index)
				)
			) catch ()
	)

)


/*
 * function: flash_CallFunction
 *  call a function in flash activex object
 *
 */
fn flash_CallFunction name args:undefined =
(
	--format "flash_CallFunction -> % %\n" name args
	-- generate xml to send to flash
	local xml = stringstream ""
	format "<invoke name=\"%\" returntype=\"xml\">" name to:xml
	-- arguments
	if classOf args == Array then (
		format "<arguments>" name to:xml
		for i = 1 to args.count do (
			-- expand to covert different data types
			case classOf args[i] of (
				String:
					format "<string>%</string>" args[i] to:xml
				Integer:
					format "<number>%</number>" args[i] to:xml
			)
		)
		format "</arguments>" name to:xml
	)
	format "</invoke>" to:xml
	local xmlString = (xml as string)
	close xml

	try (
		-- now try to call the function
		ui_bipedControl.axFlashBipSelector.CallFunction(xmlString)
	) catch (
		-- error
		local exception = getCurrentException()
		print exception
	)
)


/*
 * function: flash_updateBipSelector
 *  syncs the bip selector ui with the
 *   current biped setup
 *
 */
fn flash_updateBipSelector =
(
	-- check for pony tail links
	-- biped node indexs 18 and 19
	local ponyTail1Links = getBipedLinks(18)
	local ponyTail2Links = getBipedLinks(19)

	if ponyTail1Links.count > 0 then (
		-- show pony tail 1 links
		for i = 1 to ponyTail1Links.count do (
			flash_CallFunction "showBipPart" args:#( ("L Ponytail "+i as string) )
		)
	) else (
		-- hide hierarchical
		flash_CallFunction "hideAllBipParts" args:#("L Ponytail 1")
	)


	if ponyTail2Links.count > 0 then (
		-- show pony tail 2 links
		for i = 1 to ponyTail2Links.count do (
			flash_CallFunction "showBipPart" args:#( ("R Ponytail "+i as string) )
		)
	) else (
		-- hide hierarchical
		flash_CallFunction "hideAllBipParts" args:#("R Ponytail 1")
	)

	-- check for spine links
	-- biped node indexs 9
	local spineBipParts = getBipedLinks(9)
	flash_CallFunction "showSpineLinks" args:#( (spineBipParts.count as integer) )

)


-------------------------------------------------------
-- PROPERTY LIBRARY
-------------------------------------------------------

/*
 * function: setKeyFromLibrary
 *  create a new key and set the key params
 *  from the library data
 *
 */
fn setKeyFromLibrary libraryData applyTo =
(
	-- with any bip part selected
	for obj in selection where classof obj.baseObject == Biped_Object do
	(

		case classof obj.controller of
		(
			-- BipSlave_Control
			-- Matrix3Controller
			-- the transform controller for each object
			BipSlave_Control:
			(
				-- create new key
				biped.addNewKey obj.controller slidertime
			)

			-- Vertical_Horizontal_Turn
			-- Matrix3Controller
			-- the transform controller for the root object
			Vertical_Horizontal_Turn:
			(
				local selectedTrack = obj.transform.controller.trackselection
				-- a track selection of 0 is meaningless and will not do anything!
				if selectedTrack != undefined and selectedTrack > 0 then (
					-- create new key
					biped.addNewKey obj.controller[ cnt ] slidertime
				)

			)
		)
		-- at what key time
		local keyTime = undefined
		if applyTo == #currentKey then (
			keyTime = slidertime
		)
		-- set params to key
		for param in getPropNames libraryData do (
			local val = getProperty libraryData param
			if val != undefined then (
				changeKeyParam param (val as float) keyTime
			)
		)
	)

)


/*
 * function: getPropertyLibraryItems
 *	returns an array of all *.ini files in
 *   the 'library' directory
 *
 */
fn getPropertyLibraryItems =
(
	-- finds the *.ini files
	local libraryItems = #()
	local iniFiles = getFiles (gScriptRootDir+"library\\*.ini")
	-- reads the name
	for file in iniFiles do (
		append libraryItems (getIniSetting file "LIBRARY_DATA" "name")
	)
	return libraryItems
)


/*
 * function: getPropertyLibraryData
 *  returns libraryData collection from the file
 */
fn getPropertyLibraryData file =
(
	-- read the *.ini file
	file = gScriptRootDir+"library\\"+file+".ini"
	local key = (getIniSetting file "LIBRARY_DATA")

	-- reads the key/value pairs
	local data = libraryData()
	for i in key do (
		local v = getIniSetting file "LIBRARY_DATA" i
		case i of (
			"name":
				data.name = v
			"tension":
				data.tension = v as float
			"continuity":
				data.continuity = v as float
			"bias":
				data.bias = v as float
			"easeTo":
				data.easeTo = v as float
			"easeFrom":
				data.easeFrom = v as float
			"ikBlend":
				data.ikBlend = v as float
			"ikSpace":
				data.ikSpace = v
			"ikJoinedPivot":
				data.ikJoinedPivot = v
		)
	)
	-- otherwise
	return data
)


/*
 * function: writePropertyLibraryData
 *  writes libraryData to file
 */
fn writePropertyLibraryData libraryData =
(
	-- create file
	local fs = createFile (gScriptRootDir+"library\\"+libraryData.name+".ini")
	-- write library data to file
	if fs != undefined then	(
		-- header
		format "; Worker of Biped\n" to:fs
		format "[LIBRARY_DATA]\n" to:fs
		-- key/data pairs
		for key in getPropNames libraryData do (
			local val = getProperty libraryData key
			if val != undefined then (
				format "%=%\n" (key as string) val to:fs
			)
		)
 	)
	-- close filestream
	flush fs
	close fs
)


-------------------------------------------------------
-- USER INTERFACE
-------------------------------------------------------

	-------------------------------------------------------
	-- LEFT PANEL
	-------------------------------------------------------

	/*
	 * Biped List Rollout
	 *  a list box of all bipeds in the scene
	 *
	 */
	rollout ui_bipedList "Biped List"
	(
		listbox lsb_listOfBipeds pos:[3,5] width:160 height:5 items:#()
		button  btn_reloadList  "Reload List" pos:[3,80] width:160

		-------------------------------------------------------
		-- LOGIC
		-------------------------------------------------------

		fn updateBipedList =
		(
			local bipedArray = getBipedsInScene()

			if bipedArray.count > 0 then (
				-- update list box items
				lsb_listOfBipeds.items = bipedArray
				-- enable selection tools
				ui_selectionTools.toggleSelectionTools(true)
				-- update current node
				if lsb_listOfBipeds.selection < 1 then (
					lsb_listOfBipeds.selection = 1
				)
				gBipedNode = getNodeByName lsb_listOfBipeds.items[lsb_listOfBipeds.selection]
				-- show biped
				flash_CallFunction("showBiped")
				-- update bip selector
				flash_updateBipSelector()
			) else (
				-- clear list box
				lsb_listOfBipeds.items = #()
				-- disable selection tools
				ui_selectionTools.toggleSelectionTools(false)
				-- hide biped
				flash_CallFunction("hideBiped")
			)

		)

		-------------------------------------------------------
		-- EVENTS
		-------------------------------------------------------

		on btn_reloadList pressed do
		(
			-- reload
			updateBipedList()
		)

		on lsb_listOfBipeds selected i do
		(
			-- reload
			flash_CallFunction("clearSelection")
			updateBipedList()
		)
	)


	/*
	 * Selection Tools Rollout
	 *
	 */
	rollout ui_selectionTools "Selection Tools"
	(
		button btn_selectRoot "Select Root" width:80 height:22 pos:[3,3]
		button btn_selectBody "Select Body" width:80 height:22 pos:[83,3]
		button btn_selectOpposite "Select Opposite" width:160 pos:[3,25]
		button btn_selectSymmetry "Select Symmetry" width:160 pos:[3,47]
		button btn_selectSpine "Select Spine Links" width:160 pos:[3,69]

		-------------------------------------------------------
		-- LOGIC
		-------------------------------------------------------

		fn toggleSelectionTools flag =
		(
			-- update controls
			btn_selectRoot.enabled = flag
			btn_selectBody.enabled = flag
			btn_selectOpposite.enabled = flag
			btn_selectSymmetry.enabled = flag
			btn_selectSpine.enabled = flag
		)


		-------------------------------------------------------
		-- EVENTS
		-------------------------------------------------------

		on btn_selectRoot pressed do (
			flash_CallFunction("deselectAll")
			selectRoot()
		)

		on btn_selectBody pressed do (
			flash_CallFunction("selectAll")
			selectRoot()
		)

		on btn_selectOpposite pressed do (
			flash_CallFunction("selectOpposite")
		)

		on btn_selectSymmetry pressed do (
			flash_CallFunction("selectSymmetry")
		)

		on btn_selectSpine pressed do (
			flash_CallFunction("clearSelection")
			flash_CallFunction("selectSpine")
		)

	)


	/*
	 * Bip Selector
	 *  using a flash activex control allows
	 *  the user to make bip selections
	 *
	 */
	rollout ui_bipedControl "Bip Selection"
	(
		-- create adobe flash activex control
		activeXControl axFlashBipSelector "{D27CDB6E-AE6D-11cf-96B8-444553540000}" width:160 height:420 pos:[3,5] releaseOnClose:true

		local bipSelectorSWF = gScriptRootDir+"resources\\BipSelect.swf"

		-------------------------------------------------------
		-- EVENTS
		-------------------------------------------------------

		on ui_bipedControl open do
		(
			try (
				-- initialise activeX object
				axFlashBipSelector.movie = bipSelectorSWF
			) catch (
				-- retry
				axFlashBipSelector.movie = bipSelectorSWF
			)

		)

		on axFlashBipSelector FSCommand cmd args do (
			-- process command
			flash_FSCommand cmd args
		)

		on axFlashBipSelector OnReadyStateChange state do (
			-- possible states are 0=Loading, 1=Uninitialized, 2=Loaded, 3=Interactive, 4=Complete
			if state > 3 then (
				format "BipSelector OK\n"
				-- hide menu
				axFlashBipSelector.menu = false
			)
		)


	)


	-------------------------------------------------------
	-- RIGHT PANEL
	-------------------------------------------------------


	/*
	 * Edit Keys Rollout
	 *
	 */
	rollout ui_editKeys "Edit Keys"
	(
		button btn_deleteKeyAtCurrentTime "Delete Key at Current Time" width:160 height:22 align:#center
		group "Change Property"
		(
			radiobuttons rad_applyToKeys labels:#("Current Key","All Keys") columns:2 pos:[13,50]
			dropdownlist ddl_propertyName "" width:70 align:#left items:#("tension", "continuity", "bias", "easeTo", "easeFrom", "ikBlend", "ikSpace", "ikJoinedPivot")
			spinner spn_propertyValue range:[0,50,0] type:#float offset:[0,-25] align:#right width:60
			button btn_changeKeyProperty "Change Keys" width:141 height:22
		)

		-------------------------------------------------------
		-- EVENTS
		-------------------------------------------------------

		on btn_deleteKeyAtCurrentTime pressed do (
			for obj in selection do (
				deleteBipKey obj
			)
		)

		on btn_changeKeyProperty pressed do (
			local applyTo = undefined
			if rad_applyToKeys.state == 1 then (
				-- current key
				keyTime = slidertime
			)
			local param = (ddl_propertyName.items[ddl_propertyName.selection] as name)
			local toValue = (spn_propertyValue.value as float)
			changeKeyParam param toValue keyTime
		)


	)


	/*
	 * Property Library New
	 *
	 */
	rollout ui_PropertyLibraryNew " New"
	(
		label lbl_editTextName "Name" pos:[3,6]
		edittext edt_propertyName "" pos:[32,5] fieldWidth:140

		group "Properties"
		(
			checkbox use_tension "Tension" pos:[13,45]
			spinner tension_value pos:[92,45] enabled:false range:[0,50,25]

			checkbox use_continuity"Continuity" pos:[13,65]
			spinner continuity_value pos:[92,65] enabled:false range:[0,50,25]

			checkbox use_bias "Bias" pos:[13,85]
			spinner bias_value pos:[92,85] enabled:false range:[0,50,25]

			checkbox use_easeto "EaseTo" pos:[13,105]
			spinner easeto_value pos:[92,105] enabled:false range:[0,50,25]

			checkbox use_easefrom "EaseFrom" pos:[13,125]
			spinner easefrom_value pos:[92,125] enabled:false range:[0,50,25]

			checkbox use_blend "IK Blend" pos:[13,145]
			spinner blend_value pos:[92,145] enabled:false range:[0,1,0]

			checkbox use_space "IK Space" pos:[13,165]
			dropdownList space_value items:#("Body", "Object") pos:[92,165] enabled:false width:80

			checkbox use_joined "Joined Pivot" pos:[13,185]
		)

		button btn_createProperty "Create" width:172 height:30 pos:[4,210] enabled:false

		-------------------------------------------------------
		-- LOGIC
		-------------------------------------------------------

		fn getData a b =
		(
			if a.checked then return (b.value as float)
			-- otherwise
			return undefined
		)

		-------------------------------------------------------
		-- EVENTS
		-------------------------------------------------------

		-- enable/disable checkbox/spinner pairs
		on use_tension changed state do tension_value.enabled = state
		on use_continuity changed state do continuity_value.enabled = state
		on use_bias changed state do bias_value.enabled = state
		on use_easeto changed state do easeto_value.enabled = state
		on use_easefrom changed state do easefrom_value.enabled = state
		on use_blend changed state do blend_value.enabled = state
		on use_space changed state do space_value.enabled = state

		on edt_propertyName changed str do
		(
			-- disable create button if name has invalid characters
			-- needs an empty space at the end that will be trimmed later
			result = filterString (str+" ") ",.<>?/\\+*&^%$#@!;:()" splitEmptyTokens:false
			-- the name is valid if the result array count is 1
			btn_createProperty.enabled = (result.count == 1)
		)

		on btn_createProperty pressed do
		(
			local propertyName = trimright edt_propertyName.text
			if doesFileExist (gScriptRootDir+"library\\"+propertyName+".ini") then (
				-- error
				messageBox "A library present with this name already exists." title:"Error"
			) else (
				-- create library data
				local data = libraryData()

				data.name = propertyName
				data.tension = getData use_tension tension_value
				data.continuity = getData use_continuity continuity_value
				data.bias = getData use_bias bias_value
				data.easeTo = getData use_easeto easeto_value
				data.easeFrom = getData use_easefrom easefrom_value
				data.ikBlend = getData use_blend blend_value
				data.ikSpace = if use_space.checked then (space_value.items[space_value.selection]) as string
				data.ikJoinedPivot = use_joined.checked

				writePropertyLibraryData(data)
				ui_propertyLibrary.refeshRollout()

				destroyDialog ui_PropertyLibraryNew
			)
		)
	)


	/*
	 * Property Library Edit
	 *
	 */
	rollout ui_PropertyLibraryEdit " Edit"
	(
		local file = undefined

		label lbl_editTextName "Name" pos:[3,6]
		edittext edt_propertyName "" pos:[32,5] fieldWidth:140

		group "Properties"
		(
			checkbox use_tension "Tension" pos:[13,45] checked:false
			spinner tension_value pos:[92,45] enabled:false range:[0,50,25]

			checkbox use_continuity"Continuity" pos:[13,65] checked:false
			spinner continuity_value pos:[92,65] enabled:false range:[0,50,25]

			checkbox use_bias "Bias" pos:[13,85] checked:false
			spinner bias_value pos:[92,85] enabled:false range:[0,50,25]

			checkbox use_easeto "EaseTo" pos:[13,105] checked:false
			spinner easeto_value pos:[92,105] enabled:false range:[0,50,25]

			checkbox use_easefrom "EaseFrom" pos:[13,125] checked:false
			spinner easefrom_value pos:[92,125] enabled:false range:[0,50,25]

			checkbox use_blend "IK Blend" pos:[13,145] checked:false
			spinner blend_value pos:[92,145] enabled:false range:[0,1,0]

			checkbox use_space "IK Space" pos:[13,165] checked:false
			dropdownList space_value items:#("Body", "Object") pos:[92,165] enabled:false width:80

			checkbox use_joined "Joined Pivot" pos:[13,185]
		)

		button btn_saveChanges "Save Changes" width:172 height:30 pos:[4,210]


		-------------------------------------------------------
		-- LOGIC
		-------------------------------------------------------

		fn checkData param a b =
		(
			if param != undefined then (
				-- check box
				a.checked = true
				-- spinner
				b.enabled = true
				b.value = param
			)
		)

		fn getData a b =
		(
			if a.checked then return (b.value as float)
			-- otherwise
			return undefined
		)

		-------------------------------------------------------
		-- EVENTS
		-------------------------------------------------------

		-- enable/disable checkbox/spinner pairs
		on use_tension changed state do tension_value.enabled = state
		on use_continuity changed state do continuity_value.enabled = state
		on use_bias changed state do bias_value.enabled = state
		on use_easeto changed state do easeto_value.enabled = state
		on use_easefrom changed state do easefrom_value.enabled = state
		on use_blend changed state do blend_value.enabled = state
		on use_space changed state do space_value.enabled = state

		on ui_PropertyLibraryEdit open do
		(
			local ui_listbox = ui_propertyLibrary.lsb_propertyLibrary
			file = ui_listbox.items[ui_listbox.selection]
			local data = getPropertyLibraryData(file)

			edt_propertyName.text = data.name

			-- check data
			checkData data.tension use_tension tension_value
			checkData data.continuity use_continuity continuity_value
			checkData data.bias use_bias bias_value
			checkData data.easeTo use_easeto easeto_value
			checkData data.easeFrom use_easefrom easefrom_value
			checkData data.ikBlend use_blend blend_value

		)

		on btn_saveChanges pressed do
		(
			-- delete original so it can be re-created
			file = gScriptRootDir+"library\\"+file+".ini"

			if (deleteFile file) == false then (
				-- error
				messageBox "Cannot delete file: it is being used by another \nprogram or person." title:"Error"
			) else (
				-- create library data
				local data = libraryData()

				data.name = edt_propertyName.text
				data.tension = getData use_tension tension_value
				data.continuity = getData use_continuity continuity_value
				data.bias = getData use_bias bias_value
				data.easeTo = getData use_easeto easeto_value
				data.easeFrom = getData use_easefrom easefrom_value
				data.ikBlend = getData use_blend blend_value

				writePropertyLibraryData(data)
				ui_propertyLibrary.refeshRollout()
			)

			-- close edit dialog
			destroyDialog ui_PropertyLibraryEdit
		)


	)


	/*
	 * Property Library Info
	 *
	 */
	rollout ui_PropertyLibraryInfo " Information"
	(
		label property_name "" align:#left pos:[3,5]

		label use_tension "Tension" pos:[13,35] align:#left
		label tension_value "N/A" pos:[92,35] align:#left

		label use_continuity "Continuity" pos:[13,55] align:#left
		label continuity_value "N/A" pos:[92,55] align:#left

		label use_bias "Bias" pos:[13,75] align:#left
		label bias_value "N/A" pos:[92,75] align:#left

		label use_easeto "EaseTo" pos:[13,95] align:#left
		label easeto_value "N/A" pos:[92,95] align:#left

		label use_easefrom "EaseFrom" pos:[13,115] align:#left
		label easefrom_value "N/A" pos:[92,115] align:#left

		label use_blend "IK Blend" pos:[13,135] align:#left
		label blend_value "N/A" pos:[92,135] align:#left

		label use_space "IK Space" pos:[13,155] align:#left
		label space_value "N/A" pos:[92,155] align:#left

		label use_joined "Joined Pivot" pos:[13,175] align:#left
		label joined_value "No" pos:[92,175] align:#left


		-------------------------------------------------------
		-- LOGIC
		-------------------------------------------------------

		fn checkData param = (
			if param != undefined then (		
				return (param as string)
			)
			-- otherwise
			return "N/A"
		)

		-------------------------------------------------------
		-- EVENTS
		-------------------------------------------------------

		on ui_PropertyLibraryInfo open do
		(
			local ui_listbox = ui_propertyLibrary.lsb_propertyLibrary
			local file = ui_listbox.items[ui_listbox.selection]
			local data = getPropertyLibraryData(file)

			-- update labels
			property_name.text = "Information for "+checkData(data.name)
			tension_value.text = checkData(data.tension)
			continuity_value.text = checkData(data.continuity)
			bias_value.text = checkData(data.bias)
			easeto_value.text = checkData(data.easeTo)
			easefrom_value.text = checkData(data.easeFrom)
			blend_value.text = checkData(data.ikBlend)
			space_value.text = 	checkData(data.ikSpace)
			joined_value.text = if (data.ikJoinedPivot == "true") then "Yes"
		)
	)


	/*
	 * Property Library Rollout
	 *
	 */
	rollout ui_propertyLibrary "Property Library"
	(
		button btn_newProperty "New" width:40 height:22 pos:[3,5]
		button btn_loadProperty "Load" width:40 height:22 pos:[43,5]
		button btn_editProperty "Edit" width:40 height:22 pos:[83,5]
		button btn_deleteProperty "Delete" width:40 height:22 pos:[123,5]

		radiobuttons rad_applyToKeys labels:#("Current Key","All Keys") columns:2 pos:[15,30]
		listbox lsb_propertyLibrary width:160 height:6 pos:[3,50] items:#()
		button btn_setKeyFromLibrary "Set Key" width:160 height:22 pos:[3,140]

		-------------------------------------------------------
		-- LOGIC
		-------------------------------------------------------

		fn refeshRollout =
		(
			-- populate property library items
			lsb_propertyLibrary.items = getPropertyLibraryItems()
			-- enable/disable buttons based on if there are item in the list box
			btn_deleteProperty.enabled = (lsb_propertyLibrary.items.count > 0)
			btn_editProperty.enabled = (lsb_propertyLibrary.items.count > 0)
			btn_setKeyFromLibrary.enabled = (lsb_propertyLibrary.items.count > 0)
		)


		-------------------------------------------------------
		-- EVENTS
		-------------------------------------------------------

		on ui_propertyLibrary open do
		(
			-- updates lsb_propertyLibrary listbox control
			refeshRollout()
		)


		on lsb_propertyLibrary doubleClicked i do
		(
			createDialog ui_PropertyLibraryInfo width:180 height:210 modal:true style:#(#style_titlebar, #style_sysmenu)
		)


		on btn_newProperty pressed do
		(
			createDialog ui_PropertyLibraryNew width:180 height:245 modal:true style:#(#style_titlebar, #style_sysmenu)
		)


		on btn_loadProperty pressed do
		(
			-- prompt to open file
			local selectedFile = getOpenFileName "Load Property Description" types:"Property Description (*.ini)|*.ini"
			if selectedFile != undefined then (
				local newFile = gScriptRootDir+"library\\"+(getFilenameFile selectedFile)+".ini"
				-- check if file exists
				if doesFileExist newFile then (
					-- prompt user to replace file
					local response = queryBox "Would you like to replace existing file?" title:"File Exists"
					if response == false then return "file exists"
					-- otherwise delete file and continue
					deleteFile newFile
				)
				-- copy file to local directory
				copyFile selectedFile newFile
				-- updates lsb_propertyLibrary listbox control
				refeshRollout()
			)
		)


		on btn_editProperty pressed do
		(
			createDialog ui_PropertyLibraryEdit width:180 height:245 modal:true style:#(#style_titlebar, #style_sysmenu)
		)


		on btn_deleteProperty pressed do
		(
			-- find file
			local ui_listbox = ui_propertyLibrary.lsb_propertyLibrary
			local file = gScriptRootDir+"library\\"+ui_listbox.items[ui_listbox.selection]+".ini"
			-- and delete it (no confirm)
			if (deleteFile file) == false then (
				messageBox "Cannot delete file: it is being used by another \nprogram or person." title:"Error"
			)
			-- updates lsb_propertyLibrary listbox control
			refeshRollout()
		)


		on btn_setKeyFromLibrary pressed do
		(
			local ui_listbox = ui_propertyLibrary.lsb_propertyLibrary
			local file = ui_listbox.items[ui_listbox.selection]
			local data = getPropertyLibraryData(file)
			local applyTo = undefined
			if rad_applyToKeys.state == 1 then (
				applyTo = #currentKey
			) else (
				applyTo = #allKeys
			)
			setKeyFromLibrary data applyTo
		)
	)


	/*
	 * Viewport Toggle Rollout
	 *
	 */
	rollout ui_viewportToggle "Viewport Toggle"
	(
		checkbutton btn_toggleBipedProperty "Biped" width:40 height:22 pos:[3,5] tooltip:"Toggle Biped"
		checkbutton btn_toggleBonesProperty "Bones" width:40 height:22 pos:[43,5] tooltip:"Toggle Bones"
		checkbutton btn_toggleMeshProperty "Mesh" width:40 height:22 pos:[83,5] tooltip:"Toggle Mesh"
		checkbutton btn_toggleBoxProperty "Box" width:40 height:22 pos:[123,5] tooltip:"Toggle Box"

		-------------------------------------------------------
		-- EVENTS
		-------------------------------------------------------

		on btn_toggleBipedProperty changed toggle do
		(
			-- select all nodes
			local nodes = getAllBipedNodes()

			if toggle then (
				hide nodes
			) else (
				unhide nodes
			)
		)

		on btn_toggleBonesProperty changed toggle do
		(
			if toggle then (
				hide $'b_*'
			) else (
				unhide $'b_*'
			)
		)

		on btn_toggleMeshProperty changed toggle do
		(
			if toggle then (
				hide $'*mesh*'
			) else (
				unhide $'*mesh*'
			)
		)

		on btn_toggleBoxProperty changed toggle do
		(
			-- select all nodes
			local nodes = getAllBipedNodes()

			for i = 1 to nodes.count do (
				nodes[i].boxMode = toggle
			)
		)
	)


	/*
	 * About Worker Of Biped Rollout
	 * yeah! who made this?
	 *
	 */
	rollout ui_aboutWorkerOfBiped "About"
	(
		-- who made this?
		label lbl_about0 "Worker Of Biped v0.74" align:#center
		label lbl_about1 "Created by Dennis Roche" align:#center
		label lbl_about2 "Inspired by Bruce Creevey" align:#center
		label lbl_about3 ""
		label lbl_about4 "Based on Biped Worker" align:#center
		label lbl_about5 "by David Humpherys" align:#center
	)


-------------------------------------------------------
-- MAIN
-------------------------------------------------------


/*
 * Main Rollout
 *  main dialog rollout
 *
 */
rollout ui_WorkerOfBipedMain "Worker Of Biped v0.74"
(

	checkbutton cbtn_showRightPanel "Show Key Editing >>" pos:[2,2] width:180 checked:true height:22
	checkbutton cbtn_showLeftPanel  "<< Show Selection" pos:[185,2] width:180 checked:true height:22

	subrollout ui_leftPanel  width:180 height:730 pos:[2,26]
	subrollout ui_rightPanel width:180 height:730 pos:[185,26]

	-------------------------------------------------------
	-- CONSTUCTOR
	-------------------------------------------------------
	on ui_WorkerOfBipedMain open do
	(
		-- attach rollouts -> left panel
		addSubRollout ui_leftPanel  ui_bipedList
		addSubRollout ui_leftPanel  ui_selectionTools
		addSubRollout ui_leftPanel  ui_bipedControl

		-- attach rollouts -> right panel
		addSubRollout ui_rightPanel ui_editKeys
		addSubRollout ui_rightPanel ui_propertyLibrary
		addSubRollout ui_rightPanel ui_viewportToggle

		addSubRollout ui_rightPanel ui_aboutWorkerOfBiped

	)

	-------------------------------------------------------
	-- DECONSTUCTOR
	-------------------------------------------------------
	on ui_WorkerOfBipedMain close do
	(
		-- destroy
		removeSubRollout ui_leftPanel  ui_bipedList
		removeSubRollout ui_leftPanel  ui_selectionTools
		removeSubRollout ui_leftPanel  ui_bipedControl

		removeSubRollout ui_rightPanel ui_editKeys
		removeSubRollout ui_rightPanel ui_propertyLibrary
		removeSubRollout ui_rightPanel ui_viewportToggle

		removeSubRollout ui_rightPanel ui_aboutWorkerOfBiped
	)

	-------------------------------------------------------
	-- TIMER
	-------------------------------------------------------

	-- needed to delay the script so that the
	-- first call the flash ActiveX did not fail
	timer startUpTimer "startUpTimer" interval:1000

	on startUpTimer tick do (
		startUpTimer.active = false
		print "Initialise Worker Of Biped"
		ui_bipedList.updateBipedList()
	)

	-------------------------------------------------------
	-- EVENTS
	-------------------------------------------------------

	on cbtn_showRightPanel changed state do
	(
		if state then (
			ui_rightPanel.visible = true
			ui_WorkerOfBipedMain.width = 365
		) else (
			ui_rightPanel.visible = false
			ui_WorkerOfBipedMain.width = 185
		)
	)

	on cbtn_showLeftPanel changed state do
	(
		local dialogPos = getDialogPos ui_WorkerOfBipedMain

		if state then (
			setDialogPos ui_WorkerOfBipedMain [dialogPos.x-180, dialogPos.y]
			ui_leftPanel.visible = true
			cbtn_showRightPanel.visible = true
			cbtn_showLeftPanel.pos.x = 185
			ui_rightPanel.pos.x = 185
			ui_WorkerOfBipedMain.width = 365
		) else (
			setDialogPos ui_WorkerOfBipedMain [dialogPos.x+180, dialogPos.y]
			ui_leftPanel.visible = false
			cbtn_showRightPanel.visible = false
			cbtn_showLeftPanel.pos.x = 2
			ui_rightPanel.pos.x = 2
			ui_WorkerOfBipedMain.width = 185
		)
	)

)

createDialog ui_WorkerOfBipedMain width:365 height:730 modal:false style:#(#style_titlebar, #style_sysmenu, #style_minimizebox)

-------------------------------------------------------
-- END
-------------------------------------------------------